"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import numpy as np
import glob
import shutil
import matplotlib
matplotlib.use('qt5agg')
import matplotlib.pyplot as plt
import matplotlib.image as img

from . import transphire_utils as tu

def create_stack_combine_command(class2d_name, file_names, file_name, output_dir, settings, queue_com, name):
    """
    Create the command to combine BDB stacks.
    """

    command = None
    block_gpu = None
    gpu_list = None
    shell = None
    stack = None
    if 'ISAC2' in class2d_name:
        if tu.is_higher_version(class2d_name, '1.2'):
            command, stack = create_isac2_1_2_combine_command(
                class2d_name=class2d_name,
                file_names=file_names,
                file_name=file_name,
                output_dir=output_dir,
                settings=settings,
                )
            check_files = []
            block_gpu = True
            gpu_list = []
            shell = False
    else:
        message = '\n'.join([
            '{0}: Not known!'.format(settings['Copy']['Class2d']),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        IOError(message)

    assert command is not None, 'command not specified: {0}'.format(class2d_name)
    assert block_gpu is not None, 'block_gpu not specified: {0}'.format(class2d_name)
    assert gpu_list is not None, 'gpu_list not specified: {0}'.format(class2d_name)
    assert shell is not None, 'shell not specified: {0}'.format(class2d_name)
    assert stack is not None, 'stack not specified: {0}'.format(class2d_name)

    return command, check_files, block_gpu, gpu_list, shell, stack


def create_isac2_1_2_combine_command(class2d_name, file_names, file_name, output_dir, settings):
    """
    Create e2bdb.py combine command
    """
    try:
        shutil.rmtree(os.path.join(output_dir, file_name))
    except FileNotFoundError:
        pass
    file_names = [
        entry
        if not entry.endswith('.bdb')
        else
        'bdb:{0}'.format(entry.replace('EMAN2DB/', '').replace('.bdb', ''))
        for entry in file_names
        ]
    output_name = os.path.join(output_dir, file_name, file_name)
    command = []
    command.append(settings['Path']['e2bdb.py'])
    command.append(' '.join(file_names))
    command.append('--makevstack=bdb:{0}'.format(output_name))
    return ' '.join(command), 'bdb:{0}'.format(output_name)


def create_class2d_command(class2d_name, stack_name, file_name, output_dir, settings, queue_com, name):
    """
    Create the command to combine BDB stacks.
    """

    command = None
    block_gpu = None
    gpu_list = None
    shell = None
    if 'ISAC2' in class2d_name:
        if tu.is_higher_version(class2d_name, '1.2'):
            command, gpu, check_file = create_isac2_1_2_command(
                class2d_name=class2d_name,
                stack_name=stack_name,
                file_name=file_name,
                output_dir=output_dir,
                settings=settings,
                name=name,
                )
            check_files = [check_file]
            block_gpu = True
            gpu_list = gpu.split()
            shell = True
    else:
        message = '\n'.join([
            '{0}: Not known!'.format(settings['Copy']['CTF']),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        IOError(message)

    assert command is not None, 'command not specified: {0}'.format(class2d_name)
    assert block_gpu is not None, 'block_gpu not specified: {0}'.format(class2d_name)
    assert gpu_list is not None, 'gpu_list not specified: {0}'.format(class2d_name)
    assert shell is not None, 'shell not specified: {0}'.format(class2d_name)

    return command, check_files, block_gpu, gpu_list, shell


def create_isac2_1_2_command(class2d_name, stack_name, file_name, output_dir, settings, name):
    """
    Create e2bdb.py combine command
    """
    isac_output_dir = os.path.join(output_dir, file_name, 'ISAC2')
    try:
        shutil.rmtree(isac_output_dir)
    except FileNotFoundError:
        pass

    ignore_list = []
    if settings[class2d_name]['Split Gpu?'] == 'True':
        try:
            gpu_id = int(name.split('_')[-1])
        except ValueError:
            gpu_id = 0
        try:
            gpu_raw = settings[class2d_name]['--gpu_devices'].split()[gpu_id]
        except IndexError:
            raise UserWarning('There are less gpus provided than threads available! Please restart with the same number of pipeline processors as GPUs provided and restart! Stopping this thread!')
    else:
        gpu_raw = settings[class2d_name]['--gpu_devices']

    gpu = ' '.join(list(set([entry.split('_')[0] for entry in gpu_raw.split()])))
    if len(gpu.split()) != len(gpu_raw.split()) and settings[class2d_name]['Split Gpu?'] == 'False':
        raise UserWarning('One cannot use multi GPU in combination with the disabled Split GPU option!')

    command = []

    command.append("PATH=$(dirname $(which {0})):${{PATH}}".format(settings['Path'][class2d_name]))
    command.append('CUDA_VISIBLE_DEVICES={0}'.format(','.join(gpu.split())))
    command.append(settings['Path']['mpirun'])
    command.append('-np {0}'.format(settings[class2d_name]['MPI processes']))
    command.append(settings['Path'][class2d_name])
    command.append(stack_name)
    command.append(isac_output_dir)

    ignore_list.append('--CTF')
    ignore_list.append('--VPP')
    if settings[class2d_name]['--CTF'] == 'True' and settings[class2d_name]['--VPP'] == 'True':
        command.append('--VPP')
    else:
        for entry in ignore_list:
            if settings[class2d_name][entry] == 'True':
                command.append(entry)

    ignore_list.append('Nr. Particles')
    ignore_list.append('MPI processes')
    ignore_list.append('Split Gpu?')
    ignore_list.append('--gpu_devices')
    command.append('--gpu_devices={0}'.format(','.join(gpu.split())))

    for key in settings[class2d_name]:
        if key in ignore_list:
            continue
        else:
            command.append(key)
            command.append(
                '{0}'.format(settings[class2d_name][key])
                )

    command.append(';')
    command.append('mkdir -p {0}'.format(os.path.join(output_dir, file_name, 'png')))
    command.append(';')
    command.append("PATH=$(dirname $(which {0})):${{PATH}}".format(settings['Path']['e2proc2d.py']))
    command.append(settings['Path']['e2proc2d.py'])
    command.append(os.path.join(isac_output_dir, 'ordered_class_averages.hdf'))
    command.append(os.path.join(output_dir, file_name, 'png', 'ordered_class_averages.hdf.png'))
    command.append('--unstacking')
    command.append('--outmode=uint16')
    check_file = os.path.join(isac_output_dir, 'ordered_class_averages.hdf')
    return ' '.join(command), gpu, check_file


@tu.rerun_function_in_case_of_error
def create_jpg_file(file_name, output_dir):
    files = sorted(glob.glob(os.path.join(output_dir, file_name, 'png', '*')))
    dpi = 300
    if len(files) == 0:
        columns = 1
        rows = 1
        width = 1
        height = 1
    else:
        columns = np.sqrt(2*len(files))
        columns = int(columns+bool(columns % 2))
        columns += bool(columns % 2)
        rows = int(len(files) / columns + 1)

        if columns * rows < len(files):
            rows += 1
        image = img.imread(files[0])
        width = image.shape[0] * columns / dpi
        height = image.shape[1] * rows / dpi

    fig, ax = plt.subplots(rows, columns, subplot_kw={'frameon':False, 'adjustable': 'box', 'aspect': 'equal'})
    ax = np.atleast_1d(ax).ravel()
    for idx, ax_instance in enumerate(ax):
        try:
            ax_instance.imshow(img.imread(files[idx]), cmap='Greys_r')
        except IndexError:
            pass
        ax_instance.axis('off')

    fig.set_size_inches(width, height)
    aspect = width / height
    plt.subplots_adjust(wspace=0.05, hspace=0.05/aspect, top=1, bottom=0, left=0, right=1)
    tu.mkdir_p(os.path.join(output_dir, 'jpg'))
    plt.savefig(os.path.join(output_dir, 'jpg', '{0}.jpg'.format(file_name)), dpi=dpi, transparent=True, edgecolor=None) 
    plt.close(fig)


def find_logfiles(root_path, settings, queue_com, name):
    """
    Find logfiles related to the produced Extract files.

    root_path - Root path of the file
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of process

    Returns:
    list of log files
    """
    log_files = None
    copied_log_files = None
    extract_name = settings['Copy']['Class2d']
    if 'ISAC2' in extract_name:
        if tu.is_higher_version(extract_name, '1.2'):
            copied_log_files = []
            recursive_file_search(directory=root_path, files=copied_log_files)
            log_files = copied_log_files
    else:
        message = '\n'.join([
            '{0}: Not known!'.format(settings['Copy']['Class2d']),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)

    assert log_files is not None
    assert copied_log_files is not None
    return log_files, copied_log_files


def recursive_file_search(directory, files):
    """
    Recursive file search function.
    """
    file_names = glob.glob('{0}*'.format(directory))
    for name in file_names:
        if os.path.isdir(name):
            recursive_file_search('{0}/'.format(name), files)
        else:
            files.append(name)



import pwd
import sys
import os
import json
import argparse
import urllib.request
from PyQt5.QtWidgets import QApplication

import transphire
from transphire import transphire_utils as tu
from transphire.mainwindow import MainWindow
from transphire.loadwindow import DefaultSettings
from transphire import selectdialog


app = QApplication([])

a = selectdialog.SelectDialog()
a.start_retrain(
    {
        'project_folder': '/home/em-transfer-user/projects/2020_08_04_pmca_nptn_gf_krios1_superres_K3/TranSPHIRE_results',
        'log_folder': '/home/em-transfer-user/projects/2020_08_04_pmca_nptn_gf_krios1_superres_K3/TranSPHIRE_results/XXX_Log_files',
        'Path': {
            'e2proc2d.py': '/home/em-transfer-user/applications/sphire/v1.3/bin/e2proc2d.py',
            'sp_cinderella_train.py': '/home/em-transfer-user/applications/miniconda/v3.6.5/envs/cryolo_1.7.4/bin/sp_cinderella_train.py',
            'sp_cinderella_predict.py': '/home/em-transfer-user/applications/miniconda/v3.6.5/envs/cryolo_1.7.4/bin/sp_cinderella_train.py',
            },
        'Copy': {
            'Select2d': 'sp_cinderella_predict.py'
            }
        }
    )
a.enable(True)
a.show()

sys.exit(app.exec_())



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import re
import os
import sys
from PyQt5.QtWidgets import QWidget, QLabel, QFileDialog, QVBoxLayout, QComboBox, QLineEdit, QHBoxLayout, QPushButton, QAction, QStyle
from PyQt5.QtCore import pyqtSlot, pyqtSignal, Qt
from PyQt5.QtGui import QKeySequence
import subprocess

from . import transphire_utils as tu
from . import inputbox


class SettingsWidget(QWidget):
    """
    Widget for setting entrys

    Inherits:
    QWidget

    Signals:
    sig_index_changed - Emitted, if the index of a combo box changes (Combo box name|str)
    """
    sig_index_changed = pyqtSignal(str)

    def __init__(self, name, content, content_others, mount_directory, global_dict=None, input_file_names=None, parent=None):
        """
        Initialise the layout.

        Arguments:
        content - Content for the widget.
        parent - Parent widget (default None)

        Returns:
        None
        """
        super(SettingsWidget, self).__init__(parent=parent)

        self.action = QAction(self)
        self.action.setShortcut(QKeySequence(Qt.CTRL + Qt.SHIFT + Qt.Key_Return))
        self.action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.action.triggered.connect(self.enlarge)
        self.addAction(self.action)

        # Layout
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        # Global content
        self.parent = parent
        self.default = content[0]
        self.key_name = name
        self.typ = content[1]['typ']
        self.values = content[1]['values']
        self.name = content[1]['name']
        self.tooltip = content[1]['tooltip']
        self.dtype = content[1]['dtype']
        self.name_global = content[1]['name_global']
        self.global_value = 'ON-THE-FLY'
        self.widget_auto = None
        self.mount_directory = mount_directory
        try:
            dependency_name, dependency_value = content[1]['group'].split(':')
        except ValueError:
            dependency_name = None
            dependency_value = None

        if self.name == 'Project name':
            pattern = None
            pattern_example = None
            for entry in content_others:
                for widget in entry:
                    for key in widget:
                        if key == 'Project name pattern':
                            pattern = widget[key][0]
                        elif key == 'Project name pattern example':
                            pattern_example = widget[key][0]
            add_to_tooltip = [self.tooltip]
            if pattern:
                self.default = pattern
                add_to_tooltip.extend([
                    'Needs to follow:',
                    pattern,
                    ])
            if pattern_example:
                add_to_tooltip.extend([
                    'For example',
                    pattern_example
                    ])
            self.tooltip = '\n'.join(add_to_tooltip)

        if self.tooltip:
            final_tooltip = []
            for line in self.tooltip.splitlines():
                final_tooltip.append('\n'.join([line[i:i+80] for i in range(0, len(line), 80)]))
            self.tooltip = '\n'.join(final_tooltip)
        else:
            self.tooltip = self.name

        self.tooltip = '{0}\nDefault: \'{1}\''.format(self.tooltip, self.default)
        if dependency_name is not None:
            self.tooltip += '\n\nDependency: \'{0}\' is \'{1}\''.format(dependency_name, dependency_value)

        additional_widget = []
        self.add_widgets = []
        if self.typ not in ('COMBO', 'COMBOX'):
            additional_widget.append((self.enlarge, QStyle.SP_TitleBarNormalButton, 'Open the text field in "Enlarged mode".'))

        if self.typ == 'PLAIN' or self.typ == 'PASSWORD':
            self.edit = QLineEdit(self.name, self)
            self.edit.setToolTip(self.tooltip)
            self.edit.textChanged.connect(self.change_tooltip)
            self.edit.setText(self.default)

            if self.typ == 'PASSWORD':
                self.edit.setEchoMode(self.edit.Password)

            self.tooltip = '{0}\n\nShortcuts:\nCtrl + Shift + Return -> Open enlarged view'.format(self.tooltip)

        elif self.typ in ('FILE', 'FILE/SEARCH'):
            input_file_names.append(self.name)
            self.edit = QLineEdit(self.name, self)
            self.edit.textChanged.connect(self.change_tooltip)
            self.edit.setText(self.default)
            self.edit.setPlaceholderText('Press shift+return')
            self.edit.returnPressed.connect(self._find_file)
            self.tooltip = '{0}\n\nShortcuts:\nCtrl + Shift + Return -> Open file dialog\nCtrl + Return -> Open enlarged view'.format(self.tooltip)

            additional_widget.append((self._find_file, QStyle.SP_DialogOpenButton, 'Open the "Find file" dialog.'))

        elif self.typ in ('DIR', 'DIR/SEARCH'):
            self.edit = QLineEdit(self.name, self)
            self.edit.textChanged.connect(self.change_tooltip)
            self.edit.setText(self.default)
            self.edit.setPlaceholderText('Press shift+return')
            self.edit.returnPressed.connect(self._find_dir)
            self.tooltip = '{0}\n\nShortcuts:\nCtrl + Shift + Return -> Open directory dialog\nCtrl + Return -> Open enlarged view'.format(self.tooltip)

            additional_widget.append((self._find_dir, QStyle.SP_DialogOpenButton, 'Open the "Find directory" dialog.'))

        elif self.typ in ('COMBO', 'COMBOX'):
            self.edit = QComboBox(self)
            self.edit.setSizeAdjustPolicy(QComboBox.AdjustToContents)
            self.edit.currentTextChanged.connect(self.change_tooltip)
            self.edit.currentIndexChanged.connect(lambda: self.sig_index_changed.emit(self.name))
            self.edit.addItems(self.values)
            self.edit.setCurrentIndex(self.edit.findText(self.default))
            self.change_color_if_true()
            if self.typ == 'COMBOX':
                self.edit.setEditable(True)
                self.edit.setInsertPolicy(QComboBox.NoInsert)
                self.edit.editTextChanged.connect(self.change_tooltip)
                self.edit.editTextChanged.connect(lambda: self.sig_index_changed.emit(self.name))

        else:
            print('SettingsWidget:', self.typ, 'Not known! Stopping here!')
            sys.exit()

        self.label = QLabel(self.name, self)
        large_list = ['Path']
        if name in large_list:
            self.label.setObjectName('setting_large')
            self.edit.setObjectName('setting_large')
        else:
            self.label.setObjectName('setting')
            self.edit.setObjectName('setting')

        self.label.setToolTip(self.name)
        try:
            self.edit.textEdited.connect(
                lambda: self.edit.setStyleSheet(tu.get_style(typ='unchanged'))
                )
        except AttributeError:
            pass
        try:
            self.edit.currentIndexChanged.connect(
                self.change_color_if_true
                )
        except AttributeError:
            pass
        layout.addWidget(self.label)

        layout_h = QHBoxLayout()
        layout_h.setContentsMargins(0, 0, 0, 0)
        layout_h.setSpacing(0)
        layout_h.addWidget(self.edit, stretch=1)
        try:
            self.pre_global = self.edit.text()
        except AttributeError:
            self.pre_global = self.edit.currentText()

        for func, icon, tooltip in additional_widget:
            pb = QPushButton(self)
            pb.setObjectName('global')
            pb.setStyleSheet('QPushButton {color: rgba(0, 0, 0 ,0); background-color: rgba(0, 0, 0, 0)}')
            icon = pb.style().standardIcon(icon)
            pb.setIcon(icon)
            pb.setToolTip(tooltip)
            pb.clicked.connect(func)
            self.add_widgets.append(pb)

        if self.name_global is not None:
            self.edit.setObjectName('settinger')
            self.tooltip = '{0}\n\nGlobal value: {{global_value}}'.format(self.tooltip)

            self.widget_auto = QPushButton('G', self)
            self.widget_auto.setObjectName('global')
            self.widget_auto.setToolTip('Use the global value specified in the "Global" settings tab')
            state = True
            self.widget_auto.setCheckable(state)
            self.widget_auto.toggled.connect(self._toggle_change)
            self.widget_auto.setChecked(state)
            layout_h.addWidget(self.widget_auto, stretch=0)

            for global_name in self.name_global:
                if global_dict is not None:
                    global_dict.setdefault(global_name, []).append(self)

        for pb in self.add_widgets:
            layout_h.addWidget(pb, stretch=0)

        layout_h.addStretch(1)
        layout.addLayout(layout_h)

        self.tooltip = '{0}\n\nCurrent Text: \'{{current_text}}\''.format(self.tooltip)
        try:
            self.edit.textChanged.emit(self.edit.text())
        except AttributeError:
            self.edit.currentTextChanged.emit(self.edit.currentText())


    @pyqtSlot(bool)
    def _toggle_change(self, state):
        pre_pre_global = self.pre_global
        if state and self.edit.isEnabled() == True:
            try:
                self.pre_global = self.edit.text()
            except AttributeError:
                self.pre_global = self.edit.currentText()
        if self.key_name == 'Global' and state:
            try:
                current_global = self.get_current_global()
                if not current_global:
                    self.sender().setChecked(not state)
                    self.pre_global = pre_pre_global
                    return
            except Exception:
                self.pre_global = pre_pre_global
                self.sender().setChecked(not state)
                return
        elif state:
            current_global = self.global_value

        self.edit.setEnabled(not state)
        for entry in self.add_widgets:
            entry.setEnabled(not state)
        self.action.setEnabled(not state)

        if not state:
            try:
                self.edit.setText(self.pre_global)
                self.edit.setStyleSheet(tu.get_style('unchanged'))
            except AttributeError:
                self.edit.setCurrentText(self.pre_global)
                self.change_color_if_true()

        else:
            try:
                self.edit.setText(current_global)
            except AttributeError:
                self.edit.setCurrentText(current_global)
            self.edit.setStyleSheet(tu.get_style('global'))

    def get_current_global(self):
        current_global = self.pre_global
        if self.name == 'GPU':
            try:
                nvidia_output = subprocess.check_output(['nvidia-smi', '-L'])
                gpu_devices = re.findall(
                    '^GPU \d+: (.+) \(UUID: GPU-.*\)$',
                    nvidia_output.decode('utf-8'),
                    re.MULTILINE
                    )
            except subprocess.CalledProcessError:
                gpu_devices = []
            except Exception as e:
                print(e)
                gpu_devices = []
            if len(set(gpu_devices)) != 1:
                tu.message('The computer does have different types of GPUs available or no GPUs available or the GPU\'s crashed! In order to not make any mistakes, please specify the GPU IDs manually')
                return False
            else:
                current_global = ' '.join([str(entry) for entry in range(len(gpu_devices))])

        elif self.name in ('Memory usage', 'Memory usage large'):
            try:
                current_global = str(float(self.pre_global) / max(int(self.global_value), 1))
            except Exception:
                pass

        elif self.name == 'Pixel size bin':
            try:
                pixel_size_raw = float(self.parent.content['Pixel size'].get_settings()['Pixel size'])
                current_bin = int(self.parent.content['Bin X times'].get_settings()['Bin X times'])
                current_global = str(pixel_size_raw * current_bin)
            except Exception:
                pass

        elif self.name == 'Box size':
            try:
                current_radius = float(self.parent.content['Protein radius'].get_settings()['Protein radius'])
                good_box_size = [16, 24, 32, 36, 40, 44, 48, 52, 56, 60, 64, 72, 84, 96, 100, 104, 112, 120, 128, 132, 140, 168, 180, 192, 196, 208, 216, 220, 224, 240, 256, 260, 288, 300, 320, 352, 360, 384, 416, 440, 448, 480, 512, 540, 560, 576, 588, 600, 630, 640, 648, 672, 686, 700, 720, 750, 756, 768, 784, 800, 810, 840, 864, 882, 896, 900, 960, 972, 980, 1000, 1008, 1024]
                box_size = int(current_radius * 3)
                good_box_size.append(box_size)
                good_box_size = sorted(good_box_size)

                try:
                    current_global = str(good_box_size[good_box_size.index(box_size)+1])
                except IndexError:
                    current_global = str(box_size)
            except Exception:
                pass

        return current_global


    @pyqtSlot(str)
    def change_tooltip(self, text):
        edit = self.sender()
        if self.typ != 'PASSWORD':
            tooltip = self.tooltip.format(current_text=text, global_value=self.global_value)
            edit.setToolTip(tooltip)

    @pyqtSlot()
    def enlarge(self):
        if self.typ in ('COMBO', 'COMBOX'):
            return
        else:
            input_box = inputbox.InputBox(is_password=bool(self.typ == 'PASSWORD'), parent=self)
            input_box.setText('Enlarged view', self.name)
            input_box.setDefault(self.edit.text())
            input_box.set_type(self.typ)
            result = input_box.exec_()

            if result:
                self.edit.setText(input_box.getText())

    @pyqtSlot()
    def change_color_if_true(self):
        """
        Change the color, if the types are all true.

        Arguments:
        None

        Returns:
        None
        """
        if self.widget_auto is not None:
            if self.widget_auto.isChecked():
                self.edit.setStyleSheet(tu.get_style(typ='global'))
                return

        if self.edit.currentText() == 'False':
            self.edit.setStyleSheet(tu.get_style(typ='False'))
        else:
            self.edit.setStyleSheet(tu.get_style(typ='True'))

    @pyqtSlot()
    def _find_file(self):
        """
        Find file with an open file dialog.

        Arguments:
        None

        Returns:
        None
        """
        if '/SEARCH' in self.typ:
            current_dir = self.mount_directory
        else:
            current_dir = os.getcwd()

        in_file = QFileDialog.getOpenFileName(
            caption='Find file: {0}'.format(self.name),
            directory=current_dir,
            options=QFileDialog.DontUseNativeDialog
            )

        in_file = in_file[0]

        if in_file != '':
            self.edit.setText(in_file)

    @pyqtSlot()
    def _find_dir(self):
        """
        Find directory with an open directory dialog

        Arguments:
        None

        Returns:
        None
        """
        if '/SEARCH' in self.typ:
            current_dir = self.mount_directory
        else:
            current_dir = os.getcwd()

        in_dir = QFileDialog.getExistingDirectory(
            caption='Find directory: {0}'.format(self.name),
            directory=current_dir,
            options=QFileDialog.DontUseNativeDialog
            )
        if in_dir != '':
            if self.name == 'Project name':
                in_dir = os.path.basename(in_dir)
            self.edit.setText(in_dir)

            if '/SEARCH' in self.typ:
                self.parent.search_for_projects(in_dir)

    def get_settings(self, quiet=False):
        """
        Get the settings as dict.

        Arguments:
        quiet - True, if prints should not be shown.

        Returns:
        None, if an error occured.
        Settings as dictionary.
        """

        settings = {}
        if isinstance(self.edit, QComboBox):
            value = self.edit.currentText()

        elif isinstance(self.edit, QLineEdit):
            value = self.edit.text()

        else:
            message = 'Unreachable code! Please contact the TranSPHIRE authors!'
            tu.message(message)
            return None
        value = value.strip()

        if ' ' in value and (self.typ in ('FILE', 'DIR', 'FILE/SEARCH') or self.name in ('Rename prefix', 'Rename suffix', 'Project name')):
            self.edit.setStyleSheet(tu.get_style(typ='error'))
            message = '{0}: {1} needs to be {2}. To avoid problems later, file paths are not allowed to contain whitespaces. If this is the case, please rename the respective folders and files'.format(
                self.label.text(),
                value,
                self.dtype
                )

            if not quiet:
                tu.message(message)

            return None

        elif value:

            if tu.check_instance(value=value, typ=self.dtype):
                pass

            elif value == 'ON-THE-FLY' and self.widget_auto is not None:
                pass

            else:
                self.edit.setStyleSheet(tu.get_style(typ='error'))
                message = '{0}: {1} needs to be {2}.'.format(
                    self.label.text(),
                    value,
                    self.dtype
                    )

                if not quiet:
                    tu.message(message)

                return None

        else:
            pass

        settings[self.name] = value
        if self.widget_auto is not None:
            is_auto = self.widget_auto.isChecked()
        else:
            is_auto = None
        global_name = None if self.name_global is None else self.name_global[0]
        settings['{0}_global'.format(self.name)] = [global_name, is_auto]

        return settings

    def get_combo_entries(self):
        """
        Get the entries of the combo boxes.

        Arguments:
        None

        Returns:
        List containing entries.
        """
        entries_list = []
        for idx in range(self.edit.count()):
            entries_list.append(self.edit.itemText(idx))
        return entries_list

    def set_settings(self, text, is_checked):
        """
        Set settings

        text - Text to set to the widget.

        Returns:
        None
        """
        if is_checked is not None:
            is_checked_type = is_checked.split(', ')[1][:-1]
            if is_checked_type == 'None':
                is_checked = None
            elif is_checked_type == 'True':
                is_checked = True
            elif is_checked_type == 'False':
                is_checked = False
            else:
                assert False, is_checked_type

            if is_checked is not None:
                try:
                    self.widget_auto.setChecked(is_checked)
                except AttributeError:
                    pass

        widget_auto_checked = self.widget_auto.isChecked() if self.widget_auto is not None else False

        if is_checked or widget_auto_checked:
            self.pre_global = text
        else:
            if self.typ in ('COMBO', 'COMBOX'):
                index = self.edit.findText(text)
                if index == -1:
                    index = 0
                else:
                    pass
                self.edit.setCurrentIndex(index-1)
                self.edit.setCurrentIndex(index)
            else:
                self.edit.setText(text)



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5.QtWidgets import QLabel, QWidget, QHBoxLayout, QVBoxLayout, QPushButton



class Separator(QWidget):
    """Separator widget"""

    def __init__(
            self,
            typ,
            color,
            up=False,
            down=False,
            left=False,
            right=False,
            parent=None
            ):
        """
        Init function for the Seperator

        Arguments:
        typ - Either horizontal or vertical
        color - Color of the widget
        up - Affects the widgets to the top
        down - Affects the widgets to the down
        left - Affects the widgets to the left
        right - Affects the widgets to the right
        parent - Parent widget
        """
        super(Separator, self).__init__(parent)

        up_char = '\u2227'
        down_char = '\u2228'
        left_char = '<'
        right_char = '>'

        self.direction_dict = {
            up_char: up,
            down_char: down,
            left_char: left,
            right_char: right,
            }

        label = QLabel()
        label.setStyleSheet('background: {0}'.format(color))
        if typ == 'horizontal':
            label.setFixedHeight(2)
            layout = QHBoxLayout(self)
            layout_2 = QVBoxLayout()
        elif typ == 'vertical':
            label.setFixedWidth(2)
            layout = QVBoxLayout(self)
            layout_2 = QHBoxLayout()
        else:
            NameError('Typ not known {0}'.format(typ))

        layout_2.setContentsMargins(0, 0, 0, 0)
        #layout_2.addStretch(1)
        layout_2.addWidget(label)
        #layout_2.addStretch(1)

        layout.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(layout_2, stretch=1)

        if up or down or left or right:
            label2 = QLabel()
            label2.setStyleSheet('background: {0}'.format(color))
            if typ == 'horizontal':
                layout_buttons = QHBoxLayout()
                layout_2_2 = QVBoxLayout()
                layout_2_3 = QVBoxLayout()
                label2.setFixedHeight(2)
            elif typ == 'vertical':
                layout_buttons = QVBoxLayout()
                layout_2_2 = QHBoxLayout()
                layout_2_3 = QHBoxLayout()
                label2.setFixedWidth(2)
            else:
                NameError('Typ not known {0}'.format(typ))

            button_list = []
            if up:
                button_list.append(QPushButton(up_char))
            if down:
                button_list.append(QPushButton(down_char))
            if left:
                button_list.append(QPushButton(left_char))
            if right:
                button_list.append(QPushButton(right_char))

            for button in button_list:
                button.toggled.connect(self.hide_show_widget)
                button.setObjectName('sep')
                button.setCheckable(True)
                if typ == 'horizontal':
                    button.setFixedHeight(15)
                elif typ == 'vertical':
                    button.setFixedWidth(15)
                else:
                    NameError('Typ not known {0}'.format(typ))
                layout_buttons.addWidget(button, stretch=0)

            layout_2_3.setContentsMargins(0, 0, 0, 0)
            layout_2_3.addStretch(1)
            layout_2_3.addLayout(layout_buttons)
            layout_2_3.addStretch(1)

            layout_2_2.setContentsMargins(0, 0, 0, 0)
            layout_2_2.addStretch(1)
            layout_2_2.addWidget(label2)
            layout_2_2.addStretch(1)

            layout.addLayout(layout_2_3, stretch=0)
            layout.addLayout(layout_2_2, stretch=1)


    def hide_show_widget(self, status):
        """
        Hide or show the widget with the status.

        Arguments:
        status - Status show or status hide.
        """
        self.direction_dict[self.sender().text()].setVisible(not status)



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import traceback as tb
import os
import multiprocessing as mp
from PyQt5.QtCore import pyqtSignal, QObject, pyqtSlot

from . import transphire_utils as tu


class PlotWorker(QObject):
    """
    Plot different information about motion correction and ctf estimation.

    Inherits:
    QObject

    Signals:
    sig_data - Emitted, if data for plotting is found (name|str, data|object, directory|str, settings|object)
    sig_notification - Emitted, if phase plate limit is reached. (text|str)
    """
    sig_data = pyqtSignal(str, str, object, str, object)
    sig_reset = pyqtSignal()
    sig_visible = pyqtSignal(bool, str)
    sig_calculate = pyqtSignal()
    sig_reset_list = pyqtSignal()
    sig_new_round = pyqtSignal()
    sig_set_visual = pyqtSignal()

    def __init__(self, parent=None):
        """
        Initialise variables.

        Arguments:
        parent - Parent widget (default None)

        Returns:
        None
        """
        super(PlotWorker, self).__init__(parent)
        self.settings = []
        self.sig_calculate.connect(self.calculate_array)
        self.sig_reset_list.connect(self.reset_list)

    @pyqtSlot()
    def reset_list(self):
        self.sig_reset.emit()
        self.settings = []

    @pyqtSlot(object)
    def set_settings(self, settings):
        """
        Set settings for the calculation of the arrays.

        name - Name of the software that calls the calculation
        directory_name - Name of the directory that contains the log files
        settings - TranSPHIRE settings

        Returns:
        None
        """
        small_settings = {'Output': {}, 'Path': {}}
        for name, name_no_feedback, directory_name, settings, current_name in settings:
            if name_no_feedback not in ('Later', 'False'):
                small_settings['Output']['Project directory']= settings['Output']['Project directory']
                small_settings['Path']['chimerax']= settings['Path']['chimerax']
                small_settings['copy_to_work_folder_feedback_0']= settings['copy_to_work_folder_feedback_0']
                small_settings['Output']['Rename prefix'] = settings['Output']['Rename prefix']
                small_settings['Output']['Rename suffix'] = settings['Output']['Rename suffix']
                small_settings['Output']['Rename micrographs'] = settings['Output']['Rename micrographs']
                self.settings.append([name, name_no_feedback, directory_name, small_settings, current_name])

        self.calculate_array()
        self.sig_set_visual.emit()

    def send_data(self, data):
        for entry in data:
            if entry is not None:
                self.sig_data.emit(*entry)

    @pyqtSlot()
    def calculate_array(self):
        """
        Calculate array.

        Returns:
        None
        """
        valid_entries = []
        for entry in self.settings[:]:
            name, name_no_feedback, directory_name, settings, current_name = entry
            if os.path.isdir(directory_name):
                valid_entries.append([name, name_no_feedback, directory_name, settings])
                self.sig_visible.emit(True, name)
            else:
                self.sig_visible.emit(False, name)
            if name_no_feedback != current_name:
                self.settings.remove(entry)

        if valid_entries:
            try:
                with mp.Pool(min(len(valid_entries), len(valid_entries))) as p:
                    data = p.starmap(self.calculate_array_now, valid_entries)
            except Exception as e:
                print(e)
                print(tb.format_exc())
            else:
                self.send_data(data)
        self.sig_new_round.emit()

    @staticmethod
    def calculate_array_now(name, name_no_feedback, directory_name, settings):
        try:
            data, _ = tu.get_function_dict()[name_no_feedback]['plot_data'](
                name=name,
                name_no_feedback=name_no_feedback,
                settings=settings,
                directory_name=directory_name
                )
        except KeyError:
            pass
        else:
            if data is None or data.size == 0:
                return None
            else:
                return [name, name_no_feedback, data, directory_name, settings]



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
import json
import pexpect as pe
import time
import os
import shutil
import re
import glob
import copy as cp
from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot
import multiprocessing as mp
import multiprocessing.managers
import queue

from . import transphire_utils as tu


class MyManager(multiprocessing.managers.BaseManager):
    pass
MyManager.register('LifoQueue', queue.LifoQueue)


class ProcessWorker(QObject):
    """
    Setup and start worker threads

    Inherits from:
    QObject

    Buttons:
    None

    Signals:
    sig_start - Connected to the run method to start the process (settings|object)
    sig_finished - Emitted, if run method finishes (No objects)
    sig_error - Emitted, if an error occured (text|str)
    sig_status - Emitted to change the status (text|str, device|str, color|str)
    sig_notification - Emitted to send a notification (text|str)
    sig_plot_ctf - Emitted to plot ctf information (ctf_name|str, ctf_settings|object, settings|object)
    sig_plot_motion - Emitted to plot motion information (motion_name|str, motion_settings|object, settings|object)
    sig_plot_picking - Emitted to plot picking information (picking_name|str, picking_settings|str, settings|object)
    """
    sig_set_project_directory = pyqtSignal(str, str, str)
    sig_start = pyqtSignal(object, object)
    sig_finished = pyqtSignal()
    sig_error = pyqtSignal(str)
    sig_status = pyqtSignal(str, object, str, str)
    sig_notification = pyqtSignal(str)
    signal_plot = pyqtSignal(object)

    def __init__(self, password, content_process, mount_directory, parent=None):
        """
        Initialize object variables.

        Arguments:
        password - Sudo password
        content_process - Pipeline content
        mount_directory - Folder containing the mount points
        parent - Parent widget (default None)

        Return:
        None
        """
        super(ProcessWorker, self).__init__(parent)
        # Variables
        self.password = password
        self.content_process = content_process
        self.mount_directory = mount_directory
        self.stop = None
        self.abort = None
        self.settings = {}
        self.idx_number = 0
        self.idx_values = 1

        # Events
        self.sig_start.connect(self.run)

    def emit_plot_signals(self, folder_list, monitor):
        settings_emit = []
        for name_no_feedback, name, prog_type, folder in self.settings['plot_emit']:
            try:
                settings_emit.append([
                    name,
                    name_no_feedback,
                    folder,
                    self.settings,
                    self.settings['Copy'][prog_type],
                    ])
            except KeyError:
                pass

        self.signal_plot.emit(settings_emit)

    @pyqtSlot(object, object)
    def run(self, settings, restart_dict):
        """
        Start the process.

        Arguments:
        settings - Transphire settings

        Return:
        None
        """
        # Set settings
        self.abort = False
        self.sig_error.emit('NEW SESSION: {}'.format('Monitor' if settings['Monitor'] else 'Start'))

        content_process = cp.deepcopy(self.content_process)
        settings['copy_software_meta'] = True
        if settings['Input']['Input frames extension'] in ('dm4'):
            settings['Output extension'] = 'mrc'
        else:
            settings['Output extension'] = settings['Input']['Input frames extension']

        # Set paths
        settings['software_meta_tar'] = os.path.join(
            settings['tar_folder'], 'Software_meta.tar'
            )

        picking_name = settings['Copy']['Picking'] 
        if picking_name not in ('False', 'Later'):
            settings[picking_name]['--threshold_old'] = settings[picking_name]['--threshold']
            value = settings[picking_name]['--weights']
            try:
                if '|||' in value:
                    external_log, local_key = value.split('|||')
                    with open(settings[external_log], 'r') as read:
                        log_data = json.load(read)
                    try:
                        set_value = log_data[settings['current_set']][picking_name][local_key]['new_file']
                    except KeyError:
                        set_value = value
                else:
                    set_value = value
            except TypeError:
                set_value = value
            settings[picking_name]['--weights_old'] = set_value

        self.settings = settings
        self.sig_set_project_directory.emit(self.settings['project_folder'], self.settings['log_folder'], self.settings['error_folder'])

        manager_lifo = MyManager()
        manager_lifo.start()
        manager = mp.Manager()
        typ_dict = {}
        share_dict = {}
        bad_dict = {}
        queue_dict = {}
        full_content = []
        for entry in content_process:
            for process in entry:
                for key in process:
                    if 'WIDGETS' in key:
                        continue
                    process[key][self.idx_values]['group'], process[key][self.idx_values]['aim'] = \
                        process[key][self.idx_values]['group'].split(';')
                    process[key][self.idx_values]['aim'] = process[key][self.idx_values]['aim'].split(',')
                    share_dict[key] = manager.list()
                    bad_dict[key] = manager.list()
                    queue_dict[key] = manager_lifo.LifoQueue()
                    #queue_dict[key] = manager.Queue()
                    typ_dict[key] = manager.dict({
                        'file_number': 0,
                        'spot': False,
                        'lost_input_meta': False,
                        'lost_input_frames': False,
                        'lost_work': False,
                        'lost_backup': False,
                        'lost_hdd': False,
                        'full_transphire': False,
                        'full_work': False,
                        'full_backup': False,
                        'full_hdd': False,
                        'unknown_error': False,
                        'delay_error': False,
                        'is_error': False,
                        'queue_list_time': 0.0,
                        'tar_idx': 0,
                        'max_running': int(self.settings['Pipeline'][key]),
                        'running': 0,
                        'do_update_count': 0,
                        'queue_list': manager.list(),
                        'queue_list_lock': manager.Lock(),
                        'queue_lock': manager.Lock(),
                        'save_lock': manager.Lock(),
                        'count_lock': manager.Lock(),
                        'error_lock': manager.Lock(),
                        'bad_lock': manager.Lock(),
                        'share_lock': manager.Lock(),
                        'write_lock': manager.Lock(),
                        'spot_dict': manager.dict(self.fill_spot_dict()),
                        'settings_file': '{0}/updated_settings_{1}.txt'.format(
                            self.settings['log_folder'],
                            key
                            ),
                        'number_file': '{0}/last_filenumber_{1}.txt'.format(
                            self.settings['log_folder'],
                            key
                            ),
                        'save_file': '{0}/Queue_{1}'.format(
                            self.settings['queue_folder'], key
                            ),
                        'done_file': '{0}/Queue_{1}_done'.format(
                            self.settings['queue_folder'], key
                            ),
                        'feedback_lock_file': '{0}/Queue_{1}_feedback_lock'.format(
                            self.settings['queue_folder'], key
                            ),
                        'list_file': '{0}/Queue_{1}_list'.format(
                            self.settings['queue_folder'], key
                            ),
                        'error_file': '{0}/Queue_{1}_error'.format(
                            self.settings['error_folder'], key
                            ),
                        })

                    full_content.append([key, process[key]])

        # Queue communication dictionary
        queue_com = {
            'log': manager.Queue(),
            'status': manager.Queue(),
            'notification': manager.Queue(),
            'error': manager.Queue(),
            'info': manager.Queue(),
            }

        # Set stop variable to the return value of the pre_check
        if settings['Monitor']:
            self.stop = False
            self.emit_plot_signals([], monitor=True)
            self.run_monitor(
                typ_dict=typ_dict,
                queue_com=queue_com,
                full_content=full_content,
                )
        else:
            self.stop = bool(self.pre_check_programs())
            self.run_process(
                typ_dict=typ_dict,
                queue_com=queue_com,
                share_dict=share_dict,
                bad_dict=bad_dict,
                queue_dict=queue_dict,
                content_process=content_process,
                full_content=full_content,
                manager=manager,
                restart_dict=restart_dict,
                )

        self.sig_finished.emit()

    def run_monitor(
            self,
            typ_dict,
            queue_com,
            full_content
        ):
        """
        Run the TranSPHIRE monitor process.

        Arguments:
        typ_dict - Dictionary for the queue types
        queue_com - Dictionary for queue communication

        Returns:
        None
        """
        def check_int(number):
            try:
                int(number)
            except:
                return False
            else:
                return True

        while True:
            if self.stop:
                text = 'Not monitoring'
                color = tu.get_color('white')
            else:
                text = 'Monitoring'
                color = tu.get_color('Running')
            for entry in full_content:
                name = entry[0]
                key = '_'.join([key for key in name.split('_') if not check_int(key)])
                try:
                    with open(typ_dict[key]['save_file']) as read:
                        nr_do = len([line for line in read.readlines() if line.strip()])
                except FileNotFoundError:
                    nr_do = 0
                try:
                    with open(typ_dict[key]['done_file']) as read:
                        nr_done = len([line for line in read.readlines() if line.strip()])
                except FileNotFoundError:
                    nr_done = 0
                queue_com['status'].put([
                    text,
                    [
                        nr_do,
                        nr_done
                        ],
                    name,
                    color
                    ])
            try:
                self.check_queue(queue_com=queue_com)
            except BrokenPipeError:
                pass

            if self.stop:
                break
            else:
                time.sleep(3)

    def run_process(self,
            typ_dict,
            queue_com,
            share_dict,
            bad_dict,
            queue_dict,
            content_process,
            full_content,
            manager,
            restart_dict,
        ):
        """
        Run the TranSPHIRE process.

        Arguments:
        typ_dict - Dictionary for the queue types
        queue_com - Dictionary for queue communication

        Returns:
        None
        """
        # Set default values for folder list and thread list
        folder_list = ['stack_folder', 'meta_folder']
        use_threads_list = ['Meta', 'Find', 'Import']

        data_frame = tu.DataFrame(manager, self.settings['data_frame'])

        # Decide if one will use copy to hdd
        self.settings['Copy']['Copy_to_hdd'] = self.settings['Copy']['Copy to hdd']
        if self.settings['Copy']['Copy to hdd'] != 'False':
            if self.settings['Copy']['Copy to hdd'] == 'Later':
                pass
            else:
                try:
                    for folder in glob.glob('{0}/*'.format(self.settings['copy_to_hdd_folder_feedback_0'])):
                        if not os.path.ismount(folder):
                            try:
                                os.listdir(folder)
                            except PermissionError:
                                pass
                            else:
                                self.sig_error.emit(
                                    'HDD not mounted or not well unmounted!' +
                                    'Please remount if you want to use HDD'
                                    )
                                return None
                        else:
                            pass
                except FileNotFoundError:
                    self.sig_error.emit(
                        'HDD not mounted or not well unmounted!' +
                        'Please remount if you want to use HDD!'
                        )
                    return None
            use_threads_list.append('Copy_to_hdd')
        else:
            pass

        names = [
            entry.replace('_entries', '')
            for entry in self.settings['Copy']
            if entry.endswith('_entries') and
            entry.replace('_entries', '').replace('_', ' ') in self.settings['Copy']
            ]
        for entry in names:
            if 'copy_to_' in entry.lower():
                # Fill folder list and threads list
                for name in ['work', 'backup']:
                    short_name = 'Copy_to_{0}'.format(name)
                    long_name = 'Copy to {0}'.format(name)
                    folder_name = '{0}_folder_feedback_0'.format(short_name.lower())
                    user_name = '{0}_user'.format(short_name)
                    self.settings['Copy'][short_name] = \
                        self.settings['Copy'][long_name]

                    if self.settings['Copy'][long_name] != 'False':
                        if self.settings['Copy'][long_name] == 'Later':
                            pass
                        elif not os.path.ismount(self.settings[folder_name]):
                            try:
                                os.listdir(self.settings[folder_name])
                            except PermissionError:
                                pass
                            except FileNotFoundError:
                                self.sig_error.emit(
                                    '{0} folder {1} not mounted!'.format(
                                        name,
                                        self.settings['Copy'][long_name]
                                        )
                                    )
                                return None
                            except OSError as err:
                                if 'Required key' in str(err):
                                    self.sig_error.emit(
                                        '\n'.join([
                                            '{0} folder {1} no longer mounted! '.format(
                                                name,
                                                self.settings['Copy'][long_name]
                                                ),
                                            'Use kinit to refresh the key'
                                            ])
                                        )
                                    return None
                                else:
                                    print(
                                        '\n'.join([
                                            'Unknown Error occured!',
                                            'Please contact the TranSPHIRE authors!'
                                            ])
                                    )
                                    raise
                            else:
                                self.sig_error.emit(
                                    '{0} folder {1} not mounted!'.format(
                                        name,
                                        self.settings['Copy'][long_name]
                                        )
                                    )
                                return None

                        else:
                            pass
                        try:
                            self.settings[user_name] = self.settings[
                                'user_{0}'.format(
                                    self.settings['Copy'][long_name].replace(' ', '_')
                                    )
                                ]
                        except KeyError:
                            self.sig_error.emit('{0} needs remount! '.format(name))
                            return None
                        use_threads_list.append(short_name)
                    else:
                        pass

            elif 'ctf' == entry.lower():
                # Set CTF settings
                if self.settings['Copy']['CTF'] != 'False':
                    folder_list.append('ctf_folder_feedback_0')
                    use_threads_list.append('CTF')
                    ctf_name = self.settings['Copy']['CTF']
                    try:
                        if self.settings[ctf_name]['Use movies'] == 'True':
                            self.settings['Copy']['CTF_frames'] = 'True'
                            self.settings['Copy']['CTF_sum'] = 'False'
                        else:
                            self.settings['Copy']['CTF_frames'] = 'False'
                            self.settings['Copy']['CTF_sum'] = 'True'
                    except KeyError:
                        self.settings['Copy']['CTF_frames'] = 'False'
                        self.settings['Copy']['CTF_sum'] = 'True'
                else:
                    self.settings['Copy']['CTF_frames'] = 'False'
                    self.settings['Copy']['CTF_sum'] = 'False'
            else:
                if self.settings['Copy'][entry] != 'False':
                    folder_list.append('{0}_folder_feedback_0'.format(entry.lower()))
                    use_threads_list.append(entry)
                else:
                    pass

        # Fill different dictionarys with process information
        gpu_mutex_dict = dict([
            ('{0}_{1}'.format(idx, idx_2), [manager.RLock(), mp.Value('i', 0)])
            if idx_2 != -1
            else
            (str(idx), [manager.RLock(), mp.Value('i', 0)])
            for idx in range(10)
            for idx_2 in range(-1, 10)
            ])

        # Shared dictionary
        shared_dict = {
            'share': share_dict,
            'bad': bad_dict,
            'queue': queue_dict,
            'global_update_lock': manager.Lock(),
            'translate_lock': manager.Lock(),
            'ctf_star_lock': manager.Lock(),
            'ctf_partres_lock': manager.Lock(),
            'motion_star_lock': manager.Lock(),
            'motion_star_relion3_lock': manager.Lock(),
            'motion_txt_lock': manager.Lock(),
            'gpu_lock': gpu_mutex_dict,
            'gpu_lock_lock': manager.Lock(),
            'data_frame_lock': manager.Lock(),
            'typ': typ_dict,
            }

        keep_list = [
            self.settings['set_folder_raw'],
            self.settings['tar_folder'],
            self.settings['meta_folder'],
            self.settings['stack_folder'],
            ]
        try:
            restart_feedback = restart_dict['feedback']
        except KeyError:
            restart_feedback = False

        if restart_feedback:
            try:
                shutil.move(
                    os.path.join(self.settings['project_folder'], '000_Feedback_results'),
                    self.settings['restart_backup_folder'],
                    )
            except FileNotFoundError:
                pass
            try:
                shutil.move(
                    self.settings['feedback_file'],
                    self.settings['restart_backup_folder'],
                    )
            except FileNotFoundError:
                pass
        else:
            keep_list.append('000_Feedback_results')
            try:
                with open(self.settings['feedback_file'], 'r') as read:
                    content = read.read().strip()
                    self.settings['do_feedback_loop'] = int(content)
            except FileNotFoundError:
                pass

        self.settings['is_superres'] = mp.Value('i', 2)
        self.settings['do_feedback_loop'] = mp.Value('i', self.settings['do_feedback_loop'])
        if self.settings['do_feedback_loop'].value == 0:
            status = 'Done'
            color = 'Finished'
        else:
            status = 'Running'
            color = 'Running'
        queue_com['status'].put([
            '{0:02d}|{1:02d}'.format(
                int(self.settings['Output']['Number of feedbacks']) - self.settings['do_feedback_loop'].value,
                int(self.settings['Output']['Number of feedbacks'])
                ),
            [status],
            'Feedbacks',
            tu.get_color(color)
            ])
        with open(self.settings['feedback_file'], 'w') as write:
            write.write(str(self.settings['do_feedback_loop'].value))

        # Fill process queues
        for entry in content_process:
            for process in entry:
                for key in process:
                    if 'WIDGETS' in key:
                        continue
                    elif process[key][1]['name'].startswith('Copy_to'):
                        continue
                    self.prefill_queue(
                        shared_dict=shared_dict,
                        entry=process[key][1],
                        restart_dict=restart_dict,
                        keep_list=keep_list,
                        )
        for entry in content_process:
            for process in entry:
                for key in process:
                    if 'WIDGETS' in key:
                        continue
                    elif not process[key][1]['name'].startswith('Copy_to'):
                        continue
                    self.prefill_queue(
                        shared_dict=shared_dict,
                        entry=process[key][1],
                        restart_dict=restart_dict,
                        keep_list=keep_list,
                        )
        queue_com['info'].put('Current settings saved to: {0}'.format(self.settings['set_folder']))
        self.check_queue(queue_com=queue_com)

        # Unlock the Class2d queue in case of a TranSPHIRE crash during 2D classification
        try:
            with open(shared_dict['typ']['Class2d']['feedback_lock_file'], 'r') as read:
                in_feedback = '1' in read.read()

            if in_feedback and shared_dict['queue']['Select2d'].empty() and shared_dict['queue']['Train2d'].empty():
                with open(shared_dict['typ']['Class2d']['feedback_lock_file'], 'w') as write:
                    write.write('0')
        except FileNotFoundError:
            pass

        self.emit_plot_signals(folder_list=folder_list, monitor=False)

        # Define file number and check if file already exists
        if self.settings['Output']['Rename micrographs'] == 'True':
            try:
                with open(shared_dict['typ']['Find']['number_file'], 'r') as read:
                    shared_dict['typ']['Find']['file_number'] = int(read.readline())
            except FileNotFoundError:
                shared_dict['typ']['Find']['file_number'] = int(self.settings['Output']['Start number']) - 1
                message = '{0}: New run detected - Set start number to {1}\n'.format(
                    'Find',
                    self.settings['Output']['Start number'],
                    )
            else:
                if self.settings['Output']['Start number'] != '0' and \
                        int(self.settings['Output']['Start number']) > int(shared_dict['typ']['Find']['file_number']):
                    message = '{0}: Filenumber {1} provided and larger than last file_number {2}!\nContinue with {1}\n'.format(
                        'Find',
                        self.settings['Output']['Start number'],
                        shared_dict['typ']['Find']['file_number'],
                        )
                    shared_dict['typ']['Find']['file_number'] = int(self.settings['Output']['Start number']) - 1

                elif self.settings['Output']['Start number'] != '0' and \
                        int(self.settings['Output']['Start number']) <= int(shared_dict['typ']['Find']['file_number']):
                    self.stop = True
                    message = '{0}: Filenumber {1} provided and smaller equals the last file_number {2}!\nPlease adjust the Start number to a valid, i.e. not used yet, value or 0\n'.format(
                        'Find',
                        self.settings['Output']['Start number'],
                        shared_dict['typ']['Find']['file_number'],
                        )
                elif self.settings['Output']['Start number'] == '0':
                    message = '{0}: Continue run detected! Continue with file number: {1}'.format(
                        'Find',
                        shared_dict['typ']['Find']['file_number'],
                        )
                else:
                    self.stop = True
                    message = 'Unreachable code! Please contact the TranSPHIRE authors.'
            queue_com['error'].put(message)
            queue_com['notification'].put(message)
                

        #####
#
#        shared_dict['typ']['Find']['file_number'] = int(self.settings['Output']['Start number']) - 1
#
#        if self.settings['Output']['Rename micrographs'] == 'True':
#            new_name = '{0}/{1}{2:0{3}d}{4}'.format(
#                self.settings['meta_folder'],
#                self.settings['Output']['Rename prefix'],
#                shared_dict['typ']['Find']['file_number']+1,
#                len(self.settings['Output']['Estimated mic number']),
#                self.settings['Output']['Rename suffix']
#                )
#
#            if os.path.exists('{0}_krios_sum.mrc'.format(new_name)):
#                old_filenumber = shared_dict['typ']['Find']['file_number']
#                try:
#                    with open(shared_dict['typ']['Find']['number_file'], 'r') as read:
#                        shared_dict['typ']['Find']['file_number'] = int(read.readline())
#                except FileNotFoundError:
#                    shared_dict['typ']['Find']['file_number'] = 0
#                if self.settings['Output']['Increment number'] == 'True':
#                    message = '{0}: Filenumber {1} already exists!\n'.format(
#                        'Find',
#                        old_filenumber+1
#                        ) + \
#                        'Last one used: {0} - Continue with {1}'.format(
#                            shared_dict['typ']['Find']['file_number'],
#                            shared_dict['typ']['Find']['file_number']+1
#                            )
#                    queue_com['error'].put(message)
#                    queue_com['notification'].put(message)
#                else:
#                    self.stop = True
#                    message = '{0}: Filenumber {1} already exists!\n'.format(
#                        'Find',
#                        old_filenumber+1
#                        ) + \
#                        'Check Startnumber! Last one used: {0}'.format(shared_dict['typ']['Find']['file_number'])
#                    queue_com['error'].put(message)
#                    queue_com['notification'].put(message)
#            else:
#                pass
#        else:
#            pass

        # Start threads
        thread_list = []
        use_threads_set = set(use_threads_list)
        for key, settings_content in full_content:
            
            content_settings = settings_content[self.idx_values]
            number = self.settings['Pipeline'][key]
            names = None
            if (key == 'Find' and number != '1') or (key == 'Meta' and number != '1'):
                self.stop = True
                message = 'Find and Meta are not allowed to have more than 1 process running!'
                queue_com['error'].put(message)
                queue_com['notification'].put(message)
                names = [key]
            elif number == '1':
                names = [key]
            else:
                names = ['{0}_{1}'.format(key, idx) for idx in range(int(number))]
            assert names is not None, (key, names, number)

            for name in names:
                thread_obj = ProcessThread(
                    shared_dict=shared_dict,
                    name=name,
                    content_settings=content_settings,
                    queue_com=queue_com,
                    settings=self.settings,
                    mount_directory=self.mount_directory,
                    password=self.password,
                    use_threads_set=use_threads_set,
                    stop=mp.Value('i', self.stop),
                    abort=mp.Value('i', self.abort),
                    has_finished=mp.Value('i', 0),
                    data_frame=data_frame,
                    parent=self,
                    )
                thread = mp.Process(target=self.run_in_parallel, args=(thread_obj,))
                thread.start()
                thread_list.append([thread, name, content_settings, thread_obj])
            self.check_queue(queue_com=queue_com)
        time.sleep(1)

        # Run until the user stops the processes
        if not self.stop:
            while True:
                try:
                    self.check_queue(queue_com=queue_com)
                except BrokenPipeError:
                    pass
                if self.stop:
                    break
                else:
                    pass
                time.sleep(3)
        else:
            self.check_queue(queue_com=queue_com)

        # Indicate to stop all processes
        for key, settings_content in full_content:
            typ = settings_content[self.idx_values]['name']
            size = shared_dict['queue'][typ].qsize()
            self.sig_status.emit(
                'Stopping',
                [size, shared_dict['typ'][typ]['file_number']],
                key,
                tu.get_color('Stopped')
                )
        self.check_queue(queue_com=queue_com)

        # Send the stop signals to all threads
        for _, _, _, thread_obj in thread_list:
            thread_obj.stop.value = True
            thread_obj.abort.value = self.abort

        for _, name, _, thread_obj in thread_list:
            queue_com['log'].put('Waiting for {0} to finish!'.format(name))
            while not thread_obj.has_finished.value:
                time.sleep(1)
                self.check_queue(queue_com=queue_com)

        queue_com['status'].put([
            '{0:02d}|{1:02d}'.format(
                int(self.settings['Output']['Number of feedbacks']) - self.settings['do_feedback_loop'].value,
                int(self.settings['Output']['Number of feedbacks'])
                ),
            ['Not running'],
            'Feedbacks',
            tu.get_color('white')
            ])
        queue_com['log'].put('All done!')
        self.check_queue(queue_com=queue_com)

        final_sizes = []
        for key, settings_content in full_content:
            typ = settings_content[self.idx_values]['name']
            size = shared_dict['queue'][typ].qsize()
            final_sizes.append([key, size, typ])

        # Wait for all threads to finish
        for thread, name, _, thread_obj in thread_list:
            thread.join()
            del thread_obj
        time.sleep(0.1)

        for key, size, typ in final_sizes:
            self.sig_status.emit(
                '00|{0:02d}'.format(shared_dict['typ'][typ]['max_running']),
                [size, shared_dict['typ'][typ]['file_number']],
                key,
                tu.get_color('white')
                )
        time.sleep(1)

    @staticmethod
    def run_in_parallel(thread_obj):
        thread_obj.run()

    def pre_check_programs(self):
        """
        Check, if all programs the user wants to use are available.

        Arguments:
        None

        Return:
        True, if programs exist, else False
        """
        default_unique_types = tu.get_unique_types()
        error = False
        check_files = []
        for entry in default_unique_types:
            if entry == 'Compress':
                continue
            check_files.append(['Path', self.settings['Copy'][entry]])

        check_files.append(['Path', 'IMOD header'])

        try:
            if self.settings['Copy']['Train2d'] not in ('False', 'Later'):
                check_files.append(['Path', 'cryolo_gui.py'])
                check_files.append(['Path', 'cryolo_evaluation.py'])
                check_files.append(['Path', 'sp_pipe.py'])
            else:
                pass
        except KeyError:
            pass

        try:
            if self.settings['Copy']['Select2d'] not in ('False', 'Later'):
                check_files.append(['Path', 'sp_header.py'])
            else:
                pass
        except KeyError:
            pass

        try:
            if self.settings['Copy']['Class2d'] not in ('False', 'Later'):
                check_files.append(['Path', 'e2bdb.py'])
                check_files.append(['Path', 'mpirun'])
            else:
                pass
        except KeyError:
            pass

        try:
            if self.settings['Copy']['Auto3d'] not in ('False', 'Later'):
                check_files.append(['Path', 'sp_pipe.py'])
            else:
                pass
        except KeyError:
            pass

        try:
            if [entry for entry in map(lambda x: x not in ('False', 'Later'), ['Class2d', 'CTF', 'Extract', 'Picking', 'Select2d']) if entry]:
                check_files.append(['Path', 'e2proc2d.py'])
            else:
                pass
        except KeyError:
            pass

        if self.settings['Copy']['Compress'] == 'Compress cmd':
            check_files.append([self.settings['Copy']['Compress'], '--command_compress_path'])
            check_files.append([self.settings['Copy']['Compress'], '--command_uncompress_path'])
        else:
            pass

        for typ, name in check_files:
            if name != 'False' and name != 'Later':
                try:
                    is_file = os.path.isfile(self.settings[typ][name])
                except KeyError:
                    self.sig_error.emit(
                        '{0} path not valid or disabled (Advanced, Rare)! Please adjust it!'.format(name)
                        )
                    error = True
                else:
                    if shutil.which(self.settings[typ][name]) is not None:
                        pass
                    elif not is_file:
                        self.sig_error.emit(
                            '{0} path not valid or disabled (Advanced, Rare)! Please adjust it!'.format(name)
                            )
                        error = True
                    else:
                        pass
            else:
                pass

        ctf_name = self.settings['Copy']['CTF']
        motion_name = self.settings['Copy']['Motion']
        try:
            if self.settings[ctf_name]['Use movies'] and motion_name in ('Later', 'False'):
                self.sig_error.emit('Cannot automatically adjust the pixel size for the CTF output files in case of binned micrographs. Remember to adjust it manually if necessary.')
        except KeyError:
            pass

        train2d_name = self.settings['Copy']['Train2d']
        select2d_name = self.settings['Copy']['Select2d']
        class2d_name = self.settings['Copy']['Class2d']
        extract_name = self.settings['Copy']['Extract']
        if int(self.settings['Output']['Number of feedbacks']) != 0:
            if 'Later' in (train2d_name, select2d_name, extract_name, class2d_name):
                self.sig_error.emit('Number of feedbacks provided and Extract, Class2d, Select2d, or Train2d set to Later. Remember that particle extraction and subsequent runs will stall until a Extract, Class2d, Select2d, or Train2d  program is provided.')
            if 'False' in (train2d_name, select2d_name, extract_name, class2d_name):
                self.sig_error.emit('Number of feedbacks provided and Extract, Class2d, Select2d, or Train2d set to False. This is not possible, because the program will stall infinitely.')
                error = True

        auto3d_name = self.settings['Copy']['Auto3d']
        if auto3d_name not in ('False', 'Later'):
            if self.settings['Copy']['Class2d'] in ('False') and \
                    not self.settings[auto3d_name]['input_volume']:
                self.sig_error.emit('An input volume needs to be provided to Auto3d if no 2D classification should be performed!')
                error = True

            if self.settings[auto3d_name]['Use SSH'] == 'True':
                if self.settings['Copy']['Copy to work'] != 'False' and self.settings['Copy']['Copy to work'] != 'Later':
                    device_name = [
                        entry
                        for entry in self.settings['Mount'][self.settings['Copy']['Copy to work']]['IP'].split('/') if entry.strip()
                        ][0]
                    ssh_command = 'ssh -o "StrictHostKeyChecking no" {0}@{1} ls'.format(
                        self.settings[auto3d_name]['SSH username'],
                        device_name
                        )

                    child = pe.spawnu(ssh_command)
                    try:
                        expect_idx = child.expect(
                            [
                                "{0}@{1}'s password:".format(self.settings[auto3d_name]['SSH username'], device_name),
                                pe.EOF
                                ],
                            timeout=20
                            )
                    except pe.exceptions.TIMEOUT:
                        self.sig_error.emit('TIMEOUT: SSH "ls" command failed! Username or Password in Auto3d might be wrong or Copy to work is not consistent!!')
                        error = True

                    if error:
                        pass
                    elif expect_idx == 0 and self.settings[auto3d_name]['Need SSH password'] == 'True':
                        child.sendline(self.settings[auto3d_name]['SSH password'])
                    elif expect_idx == 1 and self.settings[auto3d_name]['Need SSH password'] == 'True':
                        self.sig_error.emit('SSH password provided, but not needed. Please check if passwordless authentification is enabled for this mount point.')
                        error = True
                    elif expect_idx == 0 and self.settings[auto3d_name]['Need SSH password'] == 'False':
                        self.sig_error.emit('SSH password expected, but none provided!')
                        error = True
                    elif expect_idx == 1 and self.settings[auto3d_name]['Need SSH password'] == 'False':
                        pass

                    try:
                        child.expect(pe.EOF, timeout=20)
                    except pe.exceptions.TIMEOUT:
                        self.sig_error.emit('TIMEOUT Password: SSH ls command failed! Username or Password in Auto3d might be wrong or Copy to work is not consistent!!')
                        error = True
                else:
                    self.sig_error.emit('"Copy to work" not specified for Auto3d ssh command.')
                    error = True
        return error

    def check_queue(self, queue_com):
        """
        Check the content of the queues and react accordingly.

        Arguments:
        queue_com

        Return:
        None
        """
        for key in queue_com:
            while not queue_com[key].empty():
                if key == 'status':
                    status, numbers, device, color = queue_com['status'].get()
                    self.sig_status.emit(status, numbers, device, color)
                elif key == 'notification':
                    notification = queue_com['notification'].get()
                    self.sig_notification.emit(notification)
                elif key == 'error':
                    error = queue_com['error'].get()
                    self.sig_error.emit(error)
                elif key == 'info':
                    error = queue_com['info'].get()
                    self.sig_error.emit(error)
                elif key == 'log':
                    log = queue_com['log'].get()
                    try:
                        with open(os.path.join(self.settings['log_folder'], 'sys_log.txt'), 'a+') as write:
                            write.write('{0}\n'.format(log))
                    except FileNotFoundError:
                        pass
                else:
                    print(
                        'Processworker - check_queue:',
                        key,
                        ' Not known!',
                        ' Unreachable code!',
                        ' Please contact the TranSPHIRE authors'
                        )

    def fill_spot_dict(self):
        """
        Fill the spot dictionary.

        Arguments:
        None

        Return:
        Spot dictionary
        """
        dictionary = {}
        spot_file = self.settings['spot_file']
        try:
            with open(spot_file, 'r') as read:
                lines = [line.rstrip() for line in read.readlines()]
            for line in lines:
                key, number = line.split('\t')
                dictionary[key] = number
        except FileNotFoundError:
            with open(spot_file, 'w') as write:
                write.write('')
        return dictionary

    def prefill_queue(self, shared_dict, entry, restart_dict, keep_list):
        """
        Prefill the queues for continue mode

        Arguments:
        shared_dict - Shared dictionary
        entry - Name of the queue process

        Return:
        None
        """
        key = entry['name']
        share = entry['group']
        shared_dict_typ = shared_dict['typ'][key]
        number_file = shared_dict_typ['number_file']
        settings_file = shared_dict_typ['settings_file']
        feedback_lock_file = shared_dict_typ['feedback_lock_file']
        save_file = shared_dict_typ['save_file']
        done_file = shared_dict_typ['done_file']
        list_file = shared_dict_typ['list_file']
        share_list = shared_dict['share'][share]
        queue = shared_dict['queue'][key]
        queue_list = shared_dict_typ['queue_list']

        if self.settings["Input"]["Software"] == "Just Stack":
            self.settings['copy_software_meta'] = False

        check_state = 0
        try:
            check_state = restart_dict[key]
        except KeyError:
            pass

        prepend_list = []
        if key.startswith('Copy_to'):
            for root, _, files in os.walk(self.settings['set_folder']):
                for entry in files:
                    prepend_list.append(os.path.join(root, entry))

        if check_state in (1, 2):
            try:
                shutil.move(
                    feedback_lock_file,
                    self.settings['restart_backup_folder'],
                    )
            except FileNotFoundError:
                pass
            try:
                shutil.move(
                    self.settings['{}_folder_feedback_{}'.format(key.lower(), self.settings['do_feedback_loop'].value)],
                    self.settings['restart_backup_folder'],
                    )
            except FileNotFoundError:
                pass
            try:
                shutil.move(
                    self.settings['{}_folder_feedback_0'.format(key.lower())],
                    self.settings['restart_backup_folder'],
                    )
            except FileNotFoundError:
                pass
            except shutil.Error:
                pass
            try:
                shutil.move(
                    number_file,
                    self.settings['restart_backup_folder'],
                    )
            except FileNotFoundError:
                pass
            if restart_dict['feedback']:
                try:
                    shutil.move(
                        settings_file,
                        self.settings['restart_backup_folder'],
                        )
                except FileNotFoundError:
                    pass

            special_cases = ('Extract', 'Train2d', 'CTF')
            lines = []
            if check_state == 2 or key in special_cases:
                for entry in (save_file, done_file):
                    if os.path.exists(entry):
                        with open(entry, 'r') as read:
                            lines.extend([line.strip() for line in read.readlines() if line.strip()])
                    else:
                        with open(entry, 'w'):
                            pass

            remove_patterns = []
            if check_state in (1, 2) and key == 'CTF':
                if restart_dict['Motion'] != 0:
                    remove_patterns = [
                        '^CTF_sum',
                        ]

            elif check_state == 1 and key == 'Extract':
                if restart_dict['Picking'] == 1 and restart_dict['CTF'] == 1:
                    # Reset Motion correction and therefore picking and CTF
                    remove_patterns = [
                        '.*',
                        ]
                elif restart_dict['Picking'] == 2 and restart_dict['CTF'] == 2:
                    # Reset Picking and CTF but not motion correction
                    remove_patterns = [
                        '.*\.box',
                        '.*partres.txt',
                        ]
                elif restart_dict['Picking'] == 1:
                    assert False, 'Picking cannot be 1 without CTF be 1'
                elif restart_dict['CTF'] == 1:
                    assert False, 'Picking cannot be 1 without CTF be 1'
                elif restart_dict['Picking'] == 2:
                    # Only reset picking
                    remove_patterns = [
                        '.*\.box',
                        ]
                elif restart_dict['CTF'] == 2:
                    # Only reset CTF
                    remove_patterns = [
                        '.*partres.txt',
                        ]

            elif check_state == 1 and key == 'Train2d':
                if restart_dict['Motion'] == 2:
                    remove_patterns = [
                        '.*'
                        ]
                else:
                    remove_patterns = [
                        '.*\.hdf'
                        ]

            for pattern in remove_patterns:
                lines = sorted([entry for entry in lines if re.search(pattern, entry) is None])

            with open(save_file, 'w') as write:
                write.write('\n'.join(sorted(lines)) + '\n')
            for line in lines:
                share_list.append(line.split('|||')[-1])
                queue.put(line)

            with open(done_file, 'w'):
                pass
            with open(list_file, 'w'):
                pass

        elif check_state == 0:
            try:
                keep_list.append(os.path.basename(self.settings['{}_folder_feedback_0'.format(key.lower())]))
            except KeyError:
                pass
            if os.path.exists(save_file):
                with open(save_file, 'r') as read:
                    lines = [line.strip() for line in read.readlines() if line.strip()]

                if key.startswith('Copy_to'):
                    if '000_Feedback_results' in keep_list:
                        keep_feedback = True
                    else:
                        keep_feedback = False

                    good_lines = []
                    for line in lines:
                        if os.path.dirname(line) == self.settings['project_folder']:
                            good_lines.append(line)
                        else:
                            for entry in keep_list:
                                if entry in line:
                                    if '000_Feedback_results' in line and not keep_feedback:
                                        pass
                                    else:
                                        good_lines.append(line)
                    lines = [entry for entry in np.unique(good_lines) if os.path.exists(entry)]
                    with open(save_file, 'w') as write:
                        write.write('\n'.join(lines) + '\n')

                for line in lines:
                    if self.settings['software_meta_tar'] in line:
                        self.settings['copy_software_meta'] = False
                    else:
                        pass
                    share_list.append(line.split('|||')[-1])
                    queue.put(line)
            else:
                with open(save_file, 'w'):
                    pass

            if os.path.exists(done_file):
                with open(done_file, 'r') as read:
                    lines = [line.strip() for line in read.readlines() if line.strip()]
                    shared_dict_typ['file_number'] = len(lines)
                for line in lines:
                    if self.settings['software_meta_tar'] in line:
                        self.settings['copy_software_meta'] = False
                    else:
                        pass
            else:
                with open(done_file, 'w'):
                    pass

            if os.path.exists(list_file):
                with open(list_file, 'r') as read:
                    lines = [line.strip() for line in read.readlines() if line.strip()]
                for line in lines:
                    if line:
                        queue_list.append(line)
                    if self.settings['software_meta_tar'] in line:
                        self.settings['copy_software_meta'] = False
                    else:
                        pass
            else:
                with open(list_file, 'w'):
                    pass

            # Tar index
            if not queue_list:
                tar_files = glob.glob(os.path.join(
                    self.settings['tar_folder'],
                    '{0}_*.tar'.format(key)
                    ))
                if tar_files:
                    shared_dict_typ['tar_idx'] = max([int(re.search('{0}_.*([0-9]{{6}})\.tar'.format(re.escape(key)), entry).group(1)) for entry in tar_files]) + 1
                else:
                    shared_dict_typ['tar_idx'] = 0
            else:
                try:
                    shared_dict_typ['tar_idx'] = max([int(re.search('{0}_.*([0-9]{{6}})\.tar'.format(re.escape(key)), entry).group(1)) for entry in queue_list if re.search('{0}_.*([0-9]{{6}})\.tar'.format(re.escape(key)), entry)])
                except ValueError:
                    pass


        if prepend_list:
            with open(save_file, 'a') as append:
                append.write('\n'.join(prepend_list) + '\n')
            for entry in prepend_list:
                share_list.append(entry.split('|||')[-1])
                queue.put(entry)

from .processthread import ProcessThread
